#ifndef __lint
static const char rcsid[] = "@(#) $Header: killdup.c,v 1.23 94/01/21 11:11:23 deyke Exp $";
#endif

#include <sys/types.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "bbs.h"
#include "md5.h"

#define SAVETIME (7*24*60*60)

#define SIZE_DIGEST 16

struct vitals {
  int ptr;
  unsigned char digest[SIZE_DIGEST];
};

static const unsigned char null_digest[SIZE_DIGEST] = {
  0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04,
  0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e
};

static FILE *fpbbs;
static const char *datafile = "killdup_data";
static const char *tempfile = "killdup_temp";
static int firstmesg;
static int hash_table[4096];
static int lastmesg;
static struct vitals *vitals;

/*---------------------------------------------------------------------------*/

#define hash_index(n) \
	((vitals[n].digest[0] << 4) | (vitals[n].digest[1] & 0x0f))

/*---------------------------------------------------------------------------*/

static void halt(const char *str)
{
  perror(str);
  exit(1);
}

/*---------------------------------------------------------------------------*/

static void getfilename(int n, char *filename)
{
  sprintf(filename,
	  "%02x/%02x/%02x/%02x",
	  (n >> 24) & 0xff,
	  (n >> 16) & 0xff,
	  (n >>  8) & 0xff,
	  (n      ) & 0xff);
}

/*---------------------------------------------------------------------------*/

static int getvitals(int n)
{

  FILE * fp;
  MD5_CTX mdContext;
  char *from;
  char *to;
  char buf[1024];
  char filename[1024];
  int chr;
  int retval = 0;

  MD5Init(&mdContext);
  getfilename(n, filename);
  if (fp = fopen(filename, "r")) {
    while (fgets(buf, sizeof(buf), fp)) {
      from = to = buf;
      while (chr = (*from++ & 0x7f))
	if (chr > ' ' && chr < 0x7f) *to++ = chr;
      *to = 0;
      if (!*buf) continue;
      if (buf[0] == 'R' && buf[1] == ':') continue;
      MD5Update(&mdContext, buf, to - buf);
    }
    fclose(fp);
    retval = 1;
  }
  MD5Final(&mdContext);
  memcpy((char *) vitals[n].digest, (char *) mdContext.digest, SIZE_DIGEST);
  return retval;
}

/*---------------------------------------------------------------------------*/

static void delete_mesg(int n)
{
  if (!fpbbs) {
    fpbbs = popen("exec bbs", "w");
    if (!fpbbs) halt("bbs");
  }
  fprintf(fpbbs, "delete %d\n", n);
  fflush(fpbbs);
}

/*---------------------------------------------------------------------------*/

int main(void)
{

  int curr;
  int fd;
  int h;
  int i;
  int prev;
  int vsize;
  long timelimit;
  struct index index;

  timelimit = time(NULL) - SAVETIME;

  if (chdir(WRKDIR)) halt(WRKDIR);

  if ((fd = open(INDEXFILE, O_RDONLY, 0644)) < 0) halt(INDEXFILE);
  if (read(fd, &index, sizeof(index)) != sizeof(index)) halt(INDEXFILE);
  if (index.date > timelimit) return 0;
  if (lseek(fd, -sizeof(index), SEEK_END) < 0) halt(INDEXFILE);
  for (; ; ) {
    if (read(fd, &index, sizeof(index)) != sizeof(index)) halt(INDEXFILE);
    if (index.date <= timelimit) {
      lastmesg = index.mesg;
      break;
    }
    if (lseek(fd, -2 * sizeof(index), SEEK_CUR) < 0) halt(INDEXFILE);
  }
  close(fd);

  vsize = (lastmesg + 1) * sizeof(*vitals);
  if (!(vitals = malloc(vsize))) halt("malloc");
  firstmesg = 1;
  if ((fd = open(datafile, O_RDONLY, 0644)) >= 0) {
    firstmesg = read(fd, vitals, vsize) / sizeof(*vitals);
    if (firstmesg < 1) firstmesg = 1;
    close(fd);
  }
  if (firstmesg > lastmesg) return 0;

  for (i = 1; i < firstmesg; i++) {
    if (memcmp((char *) vitals[i].digest, (char *) null_digest, SIZE_DIGEST)) {
      h = hash_index(i);
      vitals[i].ptr = hash_table[h];
      hash_table[h] = i;
    }
  }

  for (i = firstmesg; i <= lastmesg; i++) {
    if (getvitals(i)) {
      if (!memcmp((char *) vitals[i].digest, (char *) null_digest, SIZE_DIGEST)) {
	delete_mesg(i);
      } else {
	h = hash_index(i);
	vitals[i].ptr = hash_table[h];
	hash_table[h] = i;
	for (prev = i; curr = vitals[prev].ptr; prev = curr) {
	  if (!memcmp((char *) vitals[curr].digest, (char *) vitals[i].digest, SIZE_DIGEST)) {
	    memcpy((char *) vitals[curr].digest, (char *) null_digest, SIZE_DIGEST);
	    vitals[prev].ptr = vitals[curr].ptr;
	    delete_mesg(curr);
	    break;
	  }
	}
      }
    }
  }

  if ((fd = open(tempfile, O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0)
    halt(tempfile);
  if (write(fd, vitals, vsize) != vsize) halt(tempfile);
  close(fd);
  if (rename(tempfile, datafile)) halt(datafile);

  if (fpbbs) {
    fprintf(fpbbs, "quit\n");
    fflush(fpbbs);
    pclose(fpbbs);
  }

  return 0;
}

